/*==================================================================================

  cpVLnode - Control Point Vital Logic Node
  Version 1.6
  Jim Providenza  Santa Cruz Northern  
  SP Crossing Interlocking
  US&S Trackside Control Box
  
  cpVLnode - US&S Control Box and Model Board
    10 Inputs  
     6 Outputs
    
  IOX32
   Addr 0x20
    16 Outputs
   Addr 0x21
     6 Outputs

  10,628 code bytes
     779 ram bytes
------------------------------------------------------------------------------------
  cpNode concept and design committed on 5/31/2013 by Chuck Catania and Seth Neumann
  
  This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License. 
  To view a copy of the license, visit http://creativecommons.org/licenses/by-sa/3.0/deed.en_US

  This Kernel Sketch must be compiled with a version of the Arduino IDE less than or
  equal to 1.5.6-r2.  Later IDE versions changed the preprocessor IO pin name mapping
  for Leonardo.
------------------------------------------------------------------------------------

  Author: Chuck Catania - Model Railroad Control Systems 
  System Board:  Modern Devices BBLeo (Arduino Style Leonardo)
  
  Revision History: 
   v1.6    01/20/2015  Changed D344/D341 to display Stop when TC201/TC203 are occupied
                       Changed 14l and 14R to stop only when TC202 is occupied.
   v1.5    01/13/2015  Fixed 22R/14L aspect
                       Changed 14L to Diverging Approach when 13 set to R
                       Corrected 22R to Stop whenever any block is occupied
   v1.4    01/01/2015  Added direction of travel (DOT) for interlocking entrance and exit
           01/04/2015  Changed 22R from Diverging Clear to Diverging Approach when 22R is Reverse
   v1.3    12/28/2015  Implemented Alpha signal rules.
   v1.2    12/22/2015  Integrated and tested RSSDs, sound modeule, RSMC, and CSNK with base code 
   v1.1    11/28/2015  Input/Output bit manipulation functions coded 
                       Flashing aspects coded
   v1.0    09/22/2015  Intitial design for a cpVLnode - Control Point Vital Logic node 
      
  Reserved I/O pins for cpNode compatibility are:
   D0 - RX  CMRI RS485 Receive
   D1 - TX  CMRI RS485 Transmit
   D2 - SDA I2C Data
   D3 - SCL I2C Clock
   
                   SERIAL PORT ASSIGNMENTS           
  Serial        Bootloader/Debug monitor (USB Built in)
  
/*************************************************************/
/*********            Library Header Files           *********/
/*************************************************************/
#include <SoftwareSerial.h>
//#include <TimeLib.h>

/*==============================================*/
/*====  cpVLnode CONFIGURATION PARAMETERS   ====*/
/*==============================================*/
                                                //
                                                //      
/*==============================================*/
/*====  cpVLnode CONFIGURATION PARAMETERS   ====*/
/*==============================================*/

//----------------------
// I/O Expander boards
//----------------------
#define USE_IOX


/************************************************/
/*****        STANDARD DATA PORT MAP        *****/          
/************************************************/
const int 
// RS-485 Port
//------------
           RX      =  0,    // D0  CMRI RS485 Receive
           TX      =  1,    // D1  CMRI RS485 Transmit
           
// I2C Port
//---------
           i2C_SDA =  SDA,  // D2  SDA (data) 
           i2c_SCL =  SCL,  // D3  SCL (clock)         
                              
// Analog Pins
//------------  
           PA0     = A0,    // A0    SW13 
           PA1     = A1,    // A1    MB101        
           PA2     = A2,    // A2    MB201 
           PA3     = A3,    // A3    MB202
           PA4     = A4,    // A4    MB203
           PA5     = A5;    // A5    ONBELL

// Address the bug in the Arduino IDE which does not handle re-defined analog pins.
// Redefined analog pin names do not work in digitalWrite calls and for loop indexes.
// This has been confirmed by extensive testing.
// D12 and D13 are included in the table for ease of coding bit extraction of the high byte
//-----------------------------------------------------------------------------------------           
const byte  PD12i    = 0,  // Index to actual internal pin value to keep the compiler happy
            PD13i    = 1,
            PA0i     = 2,
            PA1i     = 3,
            PA2i     = 4,
            PA3i     = 5,
            PA4i     = 6,
            PA5i     = 7;
          
// Pin name values which work for analog pins.   PDxx included for coding convenience
//-----------------------------------------------------------------------------------
const byte numAPorts = 8;
int APortMap[numAPorts] = {12, 13, A0, A1, A2, A3, A4, A5};

boolean debugMode  = false,
        debugMode1 = false;

// Debug monitor speed  
//--------------------
const int MON_SPEED      = 1200;

const int DEBOUNCE_DELAY = 4;

// Timer variables
//----------------
unsigned long previousMS = 0;
unsigned long currentMS = 0;

/* ---------------------------------
   Flashing Signal Aspect Variables
------------------------------------*/
const byte  maxFlashing = 5;
byte        numFlashing = 0;
     
int  flashTable[3][maxFlashing];
long flashTO[maxFlashing] = {0L, 0L, 0L, 0L, 0L };

long flashRate  = 500;      // interval at which to blink signal (milliseconds)

// The TX LED is used for the heartbeat indicator.  The CMRInet interface is not
// used in this application.  Pins 4 and 5 of the MAX489 socket are jumpered
// to connect D1 to the LED.  The NE555 is also removed.
//--------------------------------------------------------------------------------------
byte heartBeat = 0;
long heartMS   = 0,
     heartRate = 750;
          
const byte maxNodeBytes =   2,
           maxNodeBits  =  16,
           
           maxIOXBytes  =  16,
           maxIOXBits   = 128;
           
           
byte       nodeInBits[maxNodeBytes],
           nodeOutBits[maxNodeBytes],
           
           IOXinBits[maxIOXBytes],
           IOXoutBits[maxIOXBytes];

byte       lastNodeInBits[maxNodeBytes],
           lastIOXinBits[maxIOXBytes];

/*------------------------------------------*/
/*   Node port name to pin name map         */
/*   -1 indicates index not assigned.  Is   */
/*   checked by the actual I/O routines.    */
/*------------------------------------------*/
         //                    index 0    1    2    3    4    5     6     7     8     9    10   11   12   13   14   15
byte       nodeNameInPort[16]  = {   4,   5,   6,   7,   8,   9,   10,   11,   12,   13,   -1,  -1,  -1,  -1,  -1,  -1 },

           nodeNameOutPort[16] = {  A0,  A1,  A2,  A3,  A4,  A5,   -1,   -1,   -1,   -1,   -1,  -1,  -1,  -1,  -1,  -1 };

/**********************************************************************/
/**********                Vital Logic Variables             **********/
/**********************************************************************/
boolean DS_MODE    = false,
        DOT_Active = false;

int ledState     = 0;      // ledState used to set the LED

const byte  _ON  = 0,      // Active low is ON
            _OFF = 1,
            
            _VAC = 1,     // Block Vacant   Active High is OFF
            _OCC = 0,     // Block Occupied
            
            _swN = 0,
            _swR = 1,
            
            _DARK    = 0,
            _GREEN   = 1,
            _YELLOW  = 2,
            _RED     = 3;
            
// cpVLnode INPUTS name index numbers
// Index corresponds to bit position in the nodeInBits array
//----------------------------------------------------------
const byte  TC101    =  0,  // Optical Detectors
            TC201    =  1,
            TC202    =  2,
            TC203    =  3,

            LEV13N   =  4,  // USS Swith Lever
            LEV13R   =  5,  

            LEV14R   =  6,  // USS Signal Lever
            LEV14C   =  7,  
            LEV14L   =  8,  

            LEVNE    =  9,  // USS Transfer Lever

// cpVLnode OUTPUTS  name index numbers
// Index corresponds to bit position in the nodeOutBits array
//-----------------------------------------------------------
            SW13     =  0,  // SWITCH 13
            
            MB101    =  1,  // Model board lamps
            MB201    =  2,  
            MB202    =  3,  
            MB203    =  4,  
            
            ONBELL   =  5,  // Annunciator bell
                      
// IOX bit number within IOX buffer OUTPUTS for IOX mapping
// Index corresponds to bit position in the IOXoutBits array
//----------------------------------------------------------
            SIG14RA  =  0,  // Green
         // SIG14RA  =  1,  // Red
            SIG14RB  =  2,  // Green
         // SIG14RB  =  3,  // Red       
            SIG14LA  =  4,  // Green
         // SIG14LA  =  5,  // Red
            SIG14LB  =  6,  // Green
         // SIG14LB  =  7,  // Red
        
            SIG20R   =  8,  // Green
         // SIG20R   =  9,  // Red
            SIG20L   = 10,  // Green
         // SIG20L   = 11,  // Red        
            SIG22RA  = 12,  // Green
         // SIG22RA  = 13,  // Red 
            SIG22RB  = 14,  // Green
         // SIG22RA  = 15,  // Red
        
            SIG22L   = 16,  // Green
         // SIG22L   = 17,  // Red
            D341     = 18,  // Green
         // D341     = 19,  // Red
            D344     = 20,  // Green
         // D344     = 21,  // Red
         
            IOXA6    = 22,  // N/C
            IOXA7    = 23,  // N/C
            
// IOX bit number within IOX buffer INPUTS for IOX mapping
// Index corresponds to bit position in the IOXinBits array
//---------------------------------------------------------
            IOXB0    =  0,  // IOX Inputs
            IOXB1    =  1,  // 
            IOXB2    =  2,  // 
            IOXB3    =  3,  //        
            IOXB4    =  4,  // 
            IOXB5    =  5,  // 
            IOXB6    =  6,  // 
            IOXB7    =  7;  // 
            
/**********************************************************************/
/**********             Phantom Train Variables              **********/
/**********************************************************************/
const int fastClockRatio  = 4,
          fastClockMinute = 60 / fastClockRatio;
           
      long StandardClock    = 0L,
           train1_StartTime = 0L,
           train2_StartTime = 0L;
           
boolean wallClockRunning    = false,
        phantomTrainRunning = false;


/************************************************
*******        Wall Clock Routines     **********
*************************************************/
void Set_Standard_Clock()
{
/*
  setTime(12, 30, 0, 1, 1, 2016);  
  StandardClock = now() + fastClockMinute;
*/
}  // SET STANDARD CLOCK

/*
void printDigits(int digits){
  // utility function for digital clock display: prints preceding colon and leading 0
  Serial.print(":");
  if(digits < 10)
    Serial.print('0');
  Serial.print(digits);
}
*/

/**********************************************************************/
/**********               Get/Set Bit Routines               **********/
/**********************************************************************/
/*---------------------------------------------------*/
/* Read a specific bit in an array of unsigned bytes */
/*---------------------------------------------------*/
static int getBit(uint8_t * data, uint16_t bit)
{
    bool value = 0;

    uint16_t byte = bit / 8;
    uint8_t offset = bit % 8;

    value = data[byte] & (1 << offset);
    
    return value;
    
}  // GET BIT


/*----------------------------------------------------*/
/* Store a specific bit in an array of unsigned bytes */
/*----------------------------------------------------*/
static void setBit(uint8_t * data, uint16_t bit, bool value)
{
    uint16_t byte = bit / 8;
    uint8_t offset = bit % 8;

    if (value)
     data[byte] |= 1 << offset;     // Store one
    else 
     data[byte] &= ~(1 << offset);  // Store zero
    
}  // SET BIT


/**********************************************************************/
/**********                      IOX VARIABLES               **********/
/**********************************************************************/
#ifdef USE_IOX

// If using the I/O expander, include the i2c library
//---------------------------------------------------
#include <Wire.h>

const byte max_IOX = 16,  // device = MCP28017 chip

// I2C command registers
//----------------------
           ACTIVELOW_REG_BASE  = 0x02,  // Port A, Port B = 0x02
           PULLUP_REG_BASE     = 0x0C,  // Port A, Port B = 0x0D
           GPIO_REG_BASE       = 0x12,  // Port A, Port B = 0x13
           
           SET_PORT_OUTPUT     = 0x00,
           SET_PORT_INPUT      = 0xFF,
           SET_PORT_PULLUPS    = 0xFF,
           SET_PORT_ACTIVELOW  = 0xFF;
           
                               // Board Addresses (duplicated for ease of indexing)   
const int IOX_devAddr[max_IOX] = {0x20,0x20, 0x21,0x21, 0x22,0x22, 0x23,0x23, 
                                  0x24,0x24, 0x25,0x25, 0x26,0x26, 0x27,0x27  },

// I2C port direction bytes 0 = OUTPUT  1 = INPUT  -1 = Not Assigned
//------------------------------------------------------------------
                             // Addr  0x20     0x21     0x22     0x23     0x24     0x25     0x26     0x27
                             // Port  A  B     A  B     A  B     A  B     A  B     A  B     A  B     A  B
          IOX_ioMap[max_IOX]   =   {  0, 0,    0, 1,   -1,-1,   -1,-1,   -1,-1,   -1,-1,   -1,-1,   -1,-1 };

// Data byte mapping table created from devAddr/ioMap configuration
// Created by Configure_IOX_Ports

// [port index][0] is I2C chip address
// [port index][1] is I2C port 0=A, 1=B
//-----------------------------------------------------------------
const byte I2C_ADDR = 0,
           I2C_PORT = 1;
          
int IOX_Output_Map[max_IOX][2],
    IOX_Input_Map[max_IOX][2];
    
// Actual count of defined I2C input and output ports
//---------------------------------------------------
byte numIOX_OUT= 0,
     numIOX_IN = 0;
    
/************************************************************************* 
***********                IOX SUPPORT ROUTINES                 ********** 
**************************************************************************
  For each defined IOX board, setup the two ports for direction.
  Input ports will have the weak pull up enabled.
  
  Packing and Unpacking routines will place data bytes sequentially in
  the IOX buffers in the order of device address.    
**************************************************************************/

//------------------------------------------------------------------------
// Collect the input/output byte assignments for the number of IOX's
// Build the tables used for processing the data bytes
//------------------------------------------------------------------------
void Configure_IOX_Ports()
{
  byte i=0,
       j,k;

// Count the number of defined ports by direction.  
// Setup the input and output IOX byte map for the CMRI buffers
//-------------------------------------------------------------
  while (i < max_IOX)    
  {
    for (j=0; j<2; j++)
     {
       k = i+j;
       switch (IOX_ioMap[k])
       {
         case  0:   // Outputs
          IOX_Output_Map[numIOX_OUT][I2C_ADDR] = IOX_devAddr[k];
          IOX_Output_Map[numIOX_OUT++][I2C_PORT] = j;
         break;
         case  1:   // Inputs
          IOX_Input_Map[numIOX_IN][I2C_ADDR] = IOX_devAddr[k];
          IOX_Input_Map[numIOX_IN++][I2C_PORT] = j;
         break;
         default: ; // No Port in slot
       } 
     }
    i=i+2;      
  } 
  
}  // CONFIGURE IOX PORTS


//-------------------------------------------------------------------------------
// For each of the defined OUTPUT bytes, set the register parameters for the chip
//-------------------------------------------------------------------------------
void Setup_IOX_Outputs()
{
  byte i;
  
  for (i=0; i<numIOX_OUT; i++)
  {      
    Wire.beginTransmission(IOX_Output_Map[i][I2C_ADDR]);  // Board Address
    Wire.write(IOX_Output_Map[i][I2C_PORT]);              // A or B Port
    Wire.write(SET_PORT_OUTPUT);                          // Command to set to Output
    Wire.endTransmission();   
  }
  
}  // SET UP IOX OUTPUTS


//--------------------------------------------------------------------------
// Unpack the bytes sent in the transmit message and output to the I2C ports
//--------------------------------------------------------------------------
void Write_IOX_Outputs()
{
  byte i;
      
// Set up values to assign CMRI output bytes to I2C bytes
//-------------------------------------------------------
  if (numIOX_OUT > 0)
  {
    i = 0;      // Last onboard output byte index           

// Write the bytes to the I2C chips
//---------------------------------    
   for (i=0; i<numIOX_OUT; i++)
    {  
      Wire.beginTransmission(IOX_Output_Map[i][I2C_ADDR]);        // Board Address
      Wire.write(GPIO_REG_BASE + IOX_Output_Map[i][I2C_PORT]);    // A or B Port
      Wire.write(IOXoutBits[i]);                                  // Data to send
      Wire.endTransmission();  
    } 
  }
  
}  // UNPACK IOX OUTPUTS


//-------------------------------------------------------------------------------
// For each of the defined INPUT bytes, set the register parameters for the chip
//-------------------------------------------------------------------------------
void Setup_IOX_Inputs()
{
  byte i,
       addr,port;
  
  
  for (i=0; i<numIOX_IN; i++)
  {     
    addr =  IOX_Input_Map[i][I2C_ADDR];
    port =  IOX_Input_Map[i][I2C_PORT]; 
    
// Set port to Inputs
//-------------------
    Wire.beginTransmission(addr); 
    Wire.write(port);              
    Wire.write(SET_PORT_INPUT);                   
    Wire.endTransmission();    

// Set port to use weak pullups
//-----------------------------
    Wire.beginTransmission(addr);         
    Wire.write(PULLUP_REG_BASE + port);  
    Wire.write(SET_PORT_PULLUPS);                              
    Wire.endTransmission();    
    
// Set port to active LOW polarity
//--------------------------------
    Wire.beginTransmission(addr);            
    Wire.write(ACTIVELOW_REG_BASE + port); 
    Wire.write(SET_PORT_ACTIVELOW);                               
    Wire.endTransmission();   
  }
    
}  // SET UP IOX INPUTS
  

//-----------------------------------------------------------------------------------
// Read the input bytes from the I2C ports and collect them into the I2C input buffer
//-----------------------------------------------------------------------------------
void Read_IOX_Inputs()
{
  byte i;
      
// Transfer the input bytes to the I2C buffer
//-------------------------------------------    
  if (numIOX_IN > 0)
  {
   for (i=0; i<numIOX_IN; i++)
    {  
      Wire.beginTransmission(IOX_Input_Map[i][I2C_ADDR]);        // Board Address
      Wire.write(GPIO_REG_BASE + IOX_Input_Map[i][I2C_PORT]);    // A or B Port
      Wire.endTransmission();  

      Wire.requestFrom(IOX_Input_Map[i][I2C_ADDR],1);            // Data to read
      IOXinBits[i] = IOXinBits[i] | Wire.read();                 // Latch the inputs
      lastIOXinBits[i] = IOXinBits[i];                           // Store the current state
    } 
  }
  
}  // READ IOX INPUTS

#endif


/*************************************************************************/
/**********            GENERAL SUPPORT ROUTINES                 **********/
/*************************************************************************/
//-----------------------
//  Get the available ram 
//-----------------------
int freeRam()
{
  extern int __heap_start, *__brkval; 
  int v; 
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}  // FREE RAM


/*---------------------------------------------------------------------
  Blink the cpNode TX LED to show the system is alive
-----------------------------------------------------------------------*/
void HeartBeat()
{
  if(currentMS - heartMS > heartRate)
   {      
     heartMS = currentMS;           // save the last time you blinked the LED      
     heartBeat = (!heartBeat);      // if the LED is off turn it on and vice-versa:     
     digitalWrite( TX, heartBeat ); // set the LED with the ledState of the variable:     
   }
     
}  // HEART BEAT


/*************************************************************************/
/******       cpNode onboard I/O bit mapping for 16 bits            ******/
/******       Only one ioMap is active per node configuration       ******/
/*************************************************************************/

/*-------------------------------------------------------------------------
  Read the actual cpNode port based on the signal name index.
  Invert to reflect the active low state set by the PULLUP option.
  port values 0-255 max
---------------------------------------------------------------------------*/
bool ReadNodePort( byte inBit )
{
  byte pinName = nodeNameInPort[inBit];
  
  if (pinName != -1)
   return (!digitalRead( pinName) );
  
}  // READ NODE PORT


/*-------------------------------------------------------------------------
  The output routines takes and writes the state to the correct output port
  on the cpNode. 
---------------------------------------------------------------------------*/
void WriteNodePort(byte outBit)
{    
  byte pinName = nodeNameOutPort[outBit];

  if (pinName != -1)
   digitalWrite( pinName,getBit(nodeOutBits, outBit) );   

}  // WRITE NODE PORT


//--------------------------------
// Store amd write cpVL output bit
//--------------------------------
void setNodeOutput(byte outPort,byte theState )
{  

  setBit(nodeOutBits, outPort, theState);
  WriteNodePort(outPort);
  
}  // SET NODE OUTPUT


/*-------------------------------------------------------------------------
  The input routines collect the bits from the digitalRead() calls and
  assigns them to byte variables.

  The ports are read twice with a delay between for input debounce.
  ---------------------------------------------------------------------------*/
void Read_Node_Inputs()
{
  byte i;
  
  lastNodeInBits[0] = nodeInBits[0];
  lastNodeInBits[1] = nodeInBits[1];
  
  for (i=0; i<2; i++)
  {
    
    setBit(nodeInBits, TC101,  ReadNodePort(TC101) );
    setBit(nodeInBits, TC201,  ReadNodePort(TC201) );
    setBit(nodeInBits, TC202,  ReadNodePort(TC202) );
    setBit(nodeInBits, TC203,  ReadNodePort(TC203) );

    setBit(nodeInBits, LEV13N, ReadNodePort(LEV13N) );
    setBit(nodeInBits, LEV13R, ReadNodePort(LEV13R) );

    setBit(nodeInBits, LEV14R, ReadNodePort(LEV14R) );
    setBit(nodeInBits, LEV14C, ReadNodePort(LEV14C) );
    setBit(nodeInBits, LEV14L, ReadNodePort(LEV14L) );
    
    setBit(nodeInBits, LEVNE,  ReadNodePort(LEVNE)  );    

// Wait for debounce time
//-----------------------
   delay(DEBOUNCE_DELAY);
  }
  
}  // READ NODE INPUTS


//--------------------------
// Play the Annunicator Bell
//--------------------------
void Sound_ONBELL()
{
  
  setNodeOutput( ONBELL,_ON ); 
  delay(250);
  setNodeOutput( ONBELL,_OFF ); 

}  // SOUND ON BELL


//--------------------------
// Throw the onboard turnout
//--------------------------
void setSW13(int NorR)
{
  
  setNodeOutput( SW13,NorR ); 
  setBit(nodeOutBits, SW13, NorR);
  
}  // SET SW13


//------------------------------------------------------------------------------------------------
// Sound the On Bell when a train enters any of the approach blocks and the interlocking is vacant
// Set DOT_Active so bell will not sound on clearing interlocking.
//------------------------------------------------------------------------------------------------
void Approach_Bell()
{
  if (getBit(nodeInBits,TC101) && (!DOT_Active) )
    if ( !getBit(nodeInBits,TC201) && !getBit(nodeInBits,TC202) && !getBit(nodeInBits,TC203) )  
     Sound_ONBELL();
  
  if (getBit(nodeInBits,TC201) && (!DOT_Active) )
    if ( !getBit(nodeInBits,TC101) && !getBit(nodeInBits,TC202) && !getBit(nodeInBits,TC203) )  
     Sound_ONBELL();
     
  if (getBit(nodeInBits,TC203) && (!DOT_Active) )
    if ( !getBit(nodeInBits,TC101) && !getBit(nodeInBits,TC201) && !getBit(nodeInBits,TC202) )  
     Sound_ONBELL();

// Check for the interlocking being vacant
//----------------------------------------
  if ( !getBit(nodeInBits,TC101) && !getBit(nodeInBits,TC201) && !getBit(nodeInBits,TC202) && !getBit(nodeInBits,TC203) )
   DOT_Active = false;
  else
   DOT_Active = true;
   
}  // APPROACH BELL


/************************************************
*******      Signal Aspect Routines    **********
*************************************************/

//---------------------------------
//  Add Flashing  
//---------------------------------
void AddFlashing( byte sigName, byte sigAspect )
{
  byte i = 0;

  if (numFlashing == maxFlashing) return;

  do
  {
    if (flashTable[0][i] == -1)
     {
       numFlashing++;
       flashTable[0][i] = sigName;
       flashTable[1][i] = sigAspect;
       flashTable[2][i] = _ON;
       flashTO[i]       = currentMS;         
       return;
    }
    
    i++;
  } while (i < maxFlashing);
  
}  // ADD FLASHING

//---------------------------------
//  Remove Flashing
//---------------------------------
void RemoveFlashing( byte sigName )
{
  int i = 0;
  
  if (numFlashing == 0) return;
  
  while (i <  maxFlashing)
  {
    
    if (flashTable[0][i] == sigName)
    {
     flashTable[0][i] = -1;
     flashTable[1][i] = -1;
     flashTable[2][i] = _OFF;
     flashTO[i]       =  0L;
     numFlashing--;
     return;
    }
    i++;
  }
  
}  // REMOVE FLASHING


//-------------------------------------
//  Flash Aspects  IOX putput bits only
//-------------------------------------
void FlashAspects()
{
  byte i;
  if (numFlashing == 0) return;
  
  for (i=0; i<maxFlashing; i++)
  {
    if (flashTable[0][i] != -1)
    {
    if (currentMS - flashTO[i] > flashRate)
     {
       if (flashTable[2][i] == _OFF)
        setHead(IOXoutBits, flashTable[0][i], flashTable[1][i] );
       else
        setHead(IOXoutBits, flashTable[0][i], _DARK );
        
        flashTable[2][i] = (!flashTable[2][i]);
        flashTO[i]       = currentMS;
     }
    }
  }
  
}  // FLASH ASPECTS


//----------------------------------------------------------
// Signal Heads are bi-color RED/GREEN driven by RSSD
// Signals are wired with the Green port as the first port
// and the Red port as the first port + 1
// Port states are active low for ON
//----------------------------------------------------------
void setHead(uint8_t * data, byte sig, byte aspect )
{
/*
 if(debugMode)
 {
  Serial.print("sig = "); Serial.print(sig);    
  Serial.print("  sig+1 = "); Serial.print(sig+1);    
  Serial.print("  aspect = "); Serial.println(aspect);    
 }
*/
  switch(aspect)
  {
       
   case _DARK: 
    setBit(data, sig,  _OFF);
    setBit(data, sig+1,_OFF);
   break;
   
   case _GREEN: 
    setBit(data, sig,  _ON);
    setBit(data, sig+1,_OFF);
   break;
   
   case _YELLOW:
    setBit(data, sig,  _ON);
    setBit(data, sig+1,_ON);
   break;
   
   case _RED:
    setBit(data, sig,  _OFF);
    setBit(data, sig+1,_ON);
   break;
   
   default:
   break;
  }     
}  // SET HEAD


/*****************************************************************
*******              Compute Signal Aspects             **********
******************************************************************/
void ComputeSignal_22R()
{
  if (getBit(nodeInBits,TC101) || 
       getBit(nodeInBits,TC201) || 
        getBit(nodeInBits,TC202) || 
         getBit(nodeInBits,TC203))
   {
     setHead( IOXoutBits, SIG22RA,_RED ); 
     setHead( IOXoutBits, SIG22RB,_RED ); 
   }
  else
   {
     if (getBit(nodeInBits,LEV13N))
      {      
       setHead( IOXoutBits, SIG22RA,_RED ); 
       setHead( IOXoutBits, SIG22RB,_RED ); 
      }
      
    if (getBit(nodeInBits,LEV13R))
     {
       setHead( IOXoutBits, SIG22RA,_RED ); 
       setHead( IOXoutBits, SIG22RB,_YELLOW ); 
      }
   }    
}  // COMPUTE SIGNAL 22R


//--------------------
// Signal Lever 14
//--------------------
void ComputeSignal_14R()
{
  // Lever 14 at Rght
  //-----------------
  if (getBit(nodeInBits,LEV14R)) 
   {
     //  Check for clear route
     //-----------------------
     if (!getBit(nodeInBits,TC101) && 
           !getBit(nodeInBits,TC202) &&
             getBit(nodeInBits,LEV13N) ) 
      {
        setHead( IOXoutBits, SIG14RA,_GREEN ); 
        setHead( IOXoutBits, SIG14RB,_RED ); 

        setHead( IOXoutBits, SIG22RA,_RED ); 
        setHead( IOXoutBits, SIG22RB,_RED ); 
      }
     else
      {
        setHead( IOXoutBits, SIG14RA,_RED ); 
        setHead( IOXoutBits, SIG14RB,_RED ); 
      }
   }
   
}  // COMPUTE SIGNAL 14R


void ComputeSignal_14C()
{
  // Lever 14 at Stop
  //-----------------
  if (getBit(nodeInBits,LEV14C)) 
   {
     setHead( IOXoutBits, SIG14LA,_RED ); 
     setHead( IOXoutBits, SIG14LB,_RED ); 
     
     setHead( IOXoutBits, SIG14RA,_RED ); 
     setHead( IOXoutBits, SIG14RB,_RED );      
   }
}  // COMPUTE SIGNAL 14R


void ComputeSignal_14L()
{
  // Lever 14 at Left
  //-----------------
  if (getBit(nodeInBits,LEV14L)) 
   {
     //  Check for clear blocks
     //------------------------
     if (!getBit(nodeInBits,TC101) && 
           !getBit(nodeInBits,TC202) ) 
       {   
         // Set opposing signals Red
         //-------------------------
         setHead( IOXoutBits, SIG22RA,_RED ); 
         setHead( IOXoutBits, SIG22RB,_RED ); 
         
         setHead( IOXoutBits, SIG14RA,_RED ); 
         setHead( IOXoutBits, SIG14RB,_RED ); 
     
         // Set signal to Clear or Approach based on SW13 position
         //-------------------------------------------------------
         if(getBit(nodeInBits,LEV13N)) 
          {
            setHead( IOXoutBits, SIG14LA,_GREEN ); 
            setHead( IOXoutBits, SIG14LB,_RED );
          }
         else
          {
            setHead( IOXoutBits, SIG14LA,_RED ); 
            setHead( IOXoutBits, SIG14LB,_YELLOW );
            setHead( IOXoutBits, SIG22RA,_RED ); 
            setHead( IOXoutBits, SIG22RB,_RED ); 
          }
       }
      else
       {
        setHead( IOXoutBits, SIG14LA,_RED ); 
        setHead( IOXoutBits, SIG14LB,_RED ); 
       }

   }   
}  // COMPUTE SIGNAL 14L


void ComputeSignal_D341()
{
  if (getBit(nodeInBits,TC203))
   {
     setHead( IOXoutBits, D341,_RED ); 
   }
  else
   {
     setHead( IOXoutBits, D341,_YELLOW ); 
   }
}  // COMPUTE SIGNAL D341

  
void ComputeSignal_D344()
{
  if (getBit(nodeInBits,TC201))
   {
     setHead( IOXoutBits, D344,_RED ); 
   }
  else
   {
     setHead( IOXoutBits, D344,_YELLOW ); 
   }
}  // COMPUTE SIGNAL D344


/*-----------------------------------------------------------------------
  Set the vital logic devices to the initial interlocking state.
  Set DS_MODE based upon the state of LEVNE
-------------------------------------------------------------------------*/
void Initialize_Vital_Logic()
{
  
  setHead( IOXoutBits, SIG14RA,_RED ); 
  setHead( IOXoutBits, SIG14RB,_RED );
  setHead( IOXoutBits, SIG14LA,_RED ); 
  setHead( IOXoutBits, SIG14LB,_RED );
  
  setHead( IOXoutBits, SIG22RA,_RED ); 
  setHead( IOXoutBits, SIG22RB,_RED );
  
  setHead( IOXoutBits, SIG22L,_RED );
  
  setHead( IOXoutBits, SIG20R,_RED ); 
  setHead( IOXoutBits, SIG20L,_RED ); 
  
  DS_MODE = (!getBit(nodeInBits,LEVNE));
  
}  // INITIALIZE VITAL LOGIC


void DO_VITAL_LOGIC()
{
 
  //-----------------------------------------------------------
  //  If any input bits have changed, process lital logic rules
  //-----------------------------------------------------------
  if  ((lastNodeInBits[0] != nodeInBits[0]) || (lastNodeInBits[1] != nodeInBits[1]) )
   {

     if (getBit(nodeInBits, LEVNE))  //If true, interlocking is not in DS mode
      {
        Initialize_Vital_Logic();
//        setHead( IOXoutBits, D341,_YELLOW ); 
//        setHead( IOXoutBits, D344,_YELLOW ); 
        heartRate = 250;
      } 
     else
      {      
//        setHead( IOXoutBits, D341,_YELLOW ); 
//        setHead( IOXoutBits, D344,_YELLOW ); 
        heartRate = 750;     
      }
                
//------------------------------------------- 
// Set the turnout position based on Lever 13
//-------------------------------------------
     if (getBit(nodeInBits, LEV13N))
      setSW13(_swN);
     else
      setSW13(_swR);

//---------------------------------------------    
// Set the the state of the occupancy detectors
//---------------------------------------------    
     Approach_Bell();

     setNodeOutput( MB101,getBit(nodeInBits, TC101) ); 
     
     setNodeOutput( MB201,getBit(nodeInBits, TC201) );
     
     setNodeOutput( MB202,getBit(nodeInBits, TC202) ); 
     
     setNodeOutput( MB203,getBit(nodeInBits, TC203) ); 

//--------------------------------------
// Set the state the of the interlocking
//--------------------------------------
     ComputeSignal_22R();  
     ComputeSignal_14R();
     ComputeSignal_14C();
     ComputeSignal_14L();     
     ComputeSignal_D341();
     ComputeSignal_D344();
   } 
   
}  // DO VITAL LOGIC


/************************************************
*******            Setup Code          **********
*************************************************/
void setup()
{
  byte i,j;
  
  // RS-485 Port
//------------
  pinMode(RX, INPUT);          // D0  CMRI RS485 Receive    
  pinMode(TX, OUTPUT);         // D1  CMRI RS485 Transmit

// I2C Ports, actual setup done by the Wire library
//-------------------------------------------------
  pinMode(i2C_SDA, OUTPUT);    // D2  SDA (data)    
  pinMode(i2c_SCL, OUTPUT);    // D3  SCL (clock)    
  
  pinMode(4,  INPUT_PULLUP);        // TC101   
  pinMode(5,  INPUT_PULLUP);        // TC201   
  pinMode(6,  INPUT_PULLUP);        // TC202   
  pinMode(7,  INPUT_PULLUP);        // TC203  
  pinMode(8,  INPUT_PULLUP);        // LEV13N   
  pinMode(9,  INPUT_PULLUP);        // LEV13R   
  pinMode(10, INPUT_PULLUP);        // LEV14R   
  pinMode(11, INPUT_PULLUP);        // LEV14C
  pinMode(12, INPUT_PULLUP);        // LEV14L  
  pinMode(13, INPUT_PULLUP);        // LEVNE   
  
  pinMode(A0,   OUTPUT);              // SW13 
  pinMode(A1,   OUTPUT);              // MB101   
  pinMode(A2,   OUTPUT);              // MB201   
  pinMode(A3,   OUTPUT);              // MB202   
  pinMode(A4,   OUTPUT);              // MB203   
  pinMode(A5,   OUTPUT);              // ONBELL  
  
// Open the monitor port if in debug mode
//---------------------------------------
  if ((debugMode) || (debugMode1))
    {
     Serial.begin(MON_SPEED); 
     while(!Serial);
    }
  else
    Serial.end();
  delay(100);
     
//-------------------------------
// Initialize the data bit arrays
//-------------------------------
  for (i=0; i<maxNodeBytes; i++)
  {
   nodeInBits[i] = -1;
   nodeOutBits[i] = 0;
   lastNodeInBits[i] = 0;
  } 
   
  for (i=0; i<maxIOXBytes; i++)
  {
   IOXinBits[i]  = 0;
   IOXoutBits[i] = 0;
   lastIOXinBits[i] = 0;
  }

  for (i=0; i<maxFlashing; i++)
  {
    flashTable[0][i] = -1;
    flashTable[1][i] = -1;
    flashTable[2][i] = -1;
  }
  
  
// Set up I/O Expander (IOX) access if present
//--------------------------------------------
#ifdef USE_IOX
  Wire.begin();
  Configure_IOX_Ports();
  Setup_IOX_Inputs();
  Setup_IOX_Outputs();
#endif

// Set the initial state of the vital logic devices
//-------------------------------------------------
  Initialize_Vital_Logic();
  
// Set the clock
//---------------------
//  Set_Standard_Clock();
  

}  // SETUP

/**************************************************
*******          Main Loop Code          **********
***************************************************/
void loop()
{
  
  currentMS = millis();
  FlashAspects();
  HeartBeat();
  
/*  
  if (now() == StandardClock)
  { 
    StandardClock = now() + fastClockMinute;
  }  
*/


// Read the input bits 
//--------------------
   Read_Node_Inputs();   // Pack the onboard inputs
#ifdef USE_IOX
   Read_IOX_Inputs();    // Pack any inputs off the I2C bus
#endif

//------------------------------
// Execute the Interlocking code
//------------------------------
  DO_VITAL_LOGIC();
       
#ifdef USE_IOX
   Write_IOX_Outputs();
#endif


}  // LOOP


